@@ -1,20 +1,21 @@
 /*
   tre-match-approx.c - TRE approximate regex matching engine
 
-  Copyright (C) 2001-2004 Ville Laurikari <vl@iki.fi>.
+  Copyright (c) 2001-2006 Ville Laurikari <vl@iki.fi>.
 
-  This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU General Public License version 2 (June
-  1991) as published by the Free Software Foundation.
+  This library is free software; you can redistribute it and/or
+  modify it under the terms of the GNU Lesser General Public
+  License as published by the Free Software Foundation; either
+  version 2.1 of the License, or (at your option) any later version.
 
-  This program is distributed in the hope that it will be useful,
+  This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-  GNU General Public License for more details.
+  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+  Lesser General Public License for more details.
 
-  You should have received a copy of the GNU General Public License
-  along with this program; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+  You should have received a copy of the GNU Lesser General Public
+  License along with this library; if not, write to the Free Software
+  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 */
 
@@ -22,7 +23,7 @@
 #include <config.h>
 #endif /* HAVE_CONFIG_H */
 
-/* AIX requires this to be the first thing in the file.  */
+/* AIX requires this to be the first thing in the file.	 */
 #ifdef TRE_USE_ALLOCA
 #ifndef __GNUC__
 # if HAVE_ALLOCA_H
@@ -39,6 +40,9 @@ char *alloca ();
 #endif
 #endif /* TRE_USE_ALLOCA */
 
+#define __USE_STRING_INLINES
+#undef __NO_INLINE__
+
 #include <assert.h>
 #include <stdlib.h>
 #include <string.h>
@@ -61,12 +65,12 @@ char *alloca ();
 #include "regex.h"
 #include "xmalloc.h"
 
-#define TRE_M_COST      0
-#define TRE_M_NUM_INS   1
-#define TRE_M_NUM_DEL   2
+#define TRE_M_COST	0
+#define TRE_M_NUM_INS	1
+#define TRE_M_NUM_DEL	2
 #define TRE_M_NUM_SUBST 3
-#define TRE_M_NUM_ERR   4
-#define TRE_M_LAST      5
+#define TRE_M_NUM_ERR	4
+#define TRE_M_LAST	5
 
 #define TRE_M_MAX_DEPTH 3
 
@@ -102,7 +106,7 @@ tre_print_reach(const tre_tnfa_t *tnfa, tre_tnfa_approx_reach_t *reach,
       int i, j;
       if (reach[id].pos < pos)
 	continue;  /* Not reached. */
-      DPRINT(("  %03d, costs ", id));
+      DPRINT(("	 %03d, costs ", id));
       for (i = 0; i <= reach[id].depth; i++)
 	{
 	  DPRINT(("["));
@@ -116,7 +120,7 @@ tre_print_reach(const tre_tnfa_t *tnfa, tre_tnfa_approx_reach_t *reach,
 	  if (i + 1 <= reach[id].depth)
 	    DPRINT((", "));
 	}
-      DPRINT(("\n       tags "));
+      DPRINT(("\n	tags "));
       for (i = 0; i < num_tags; i++)
 	{
 	  DPRINT(("%02d", reach[id].tags[i]));
@@ -220,18 +224,19 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
   tre_char_t prev_c = 0, next_c = 0;
   const char *str_byte = string;
   int pos = -1;
-#ifdef TRE_MULTIBYTE
-  int pos_add_next = 1;
-#endif /* TRE_MULTIBYTE */
+  unsigned int pos_add_next = 1;
 #ifdef TRE_WCHAR
   const wchar_t *str_wide = string;
-#endif /* TRE_WCHAR */
 #ifdef TRE_MBSTATE
   mbstate_t mbstate;
 #endif /* !TRE_WCHAR */
+#endif /* TRE_WCHAR */
   int reg_notbol = eflags & REG_NOTBOL;
   int reg_noteol = eflags & REG_NOTEOL;
   int reg_newline = tnfa->cflags & REG_NEWLINE;
+  int str_user_end = 0;
+
+  int prev_pos;
 
   /* Compilation flags for this regexp. */
   int cflags = tnfa->cflags;
@@ -262,14 +267,17 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
   memset(&mbstate, '\0', sizeof(mbstate));
 #endif /* TRE_MBSTATE */
 
-  DPRINT(("tre_tnfa_run_approx, input type %d, len %d\n", type, len));
+  DPRINT(("tre_tnfa_run_approx, input type %d, len %d, eflags %d, "
+	  "match_tags %p\n",
+	  type, len, eflags,
+	  match_tags));
   DPRINT(("max cost %d, ins %d, del %d, subst %d\n",
 	  default_params.max_cost,
 	  default_params.cost_ins,
 	  default_params.cost_del,
 	  default_params.cost_subst));
 
-  /* Allocate memory for temporary data required for matching.  This needs to
+  /* Allocate memory for temporary data required for matching.	This needs to
      be done for every matching operation to be thread safe.  This allocates
      everything in a single large block from the stack frame using alloca()
      or with malloc() if alloca is unavailable. */
@@ -328,7 +336,9 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
   for (i = 0; i < tnfa->num_states; i++)
     reach[i].pos = reach_next[i].pos = -2;
 
+  prev_pos = pos;
   GET_NEXT_WCHAR();
+  pos = 0;
 
   while (1)
     {
@@ -361,6 +371,7 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
 		  /* Compute tag values after this transition. */
 		  for (i = 0; i < num_tags; i++)
 		    reach_next[id].tags[i] = -1;
+
 		  if (trans->tags)
 		    for (i = 0; trans->tags[i] >= 0; i++)
 		      if (trans->tags[i] < num_tags)
@@ -396,14 +407,13 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
 	 give a better match than what we already have. */
       for (id = 0; id < tnfa->num_states; id++)
 	{
-	  int depth, j;
+	  int depth;
 	  int cost, cost0;
 
-	  /* XXX - `pos - 1' does not work for mbs. */
-	  if (reach[id].pos != pos - 1)
+	  if (reach[id].pos != prev_pos)
 	    {
-	      DPRINT(("  insert: %d not reached\n", id));
-	      continue;  /* Not reached. */
+	      DPRINT(("	 insert: %d not reached\n", id));
+	      continue;	 /* Not reached. */
 	    }
 
 	  depth = reach[id].depth;
@@ -455,9 +465,9 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
 	  reach_next[id].depth = reach[id].depth;
 
 	  /* Set the costs after this transition. */
-	  for (i = 0; i <= depth; i++)
-	    for (j = 0; j < TRE_M_LAST; j++)
-	      reach_next[id].costs[i][j] = reach[id].costs[i][j];
+	  memcpy(reach_next[id].costs, reach[id].costs,
+		 sizeof(reach_next[id].costs[0][0])
+		 * TRE_M_LAST * (depth + 1));
 	  reach_next[id].costs[depth][TRE_M_COST] = cost;
 	  reach_next[id].costs[depth][TRE_M_NUM_INS]++;
 	  reach_next[id].costs[depth][TRE_M_NUM_ERR]++;
@@ -473,7 +483,7 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
 
       /* Handle deletes.  This is done by traversing through the whole TNFA
 	 pretending that all transitions are epsilon transitions, until
-         no more states can be reached with better costs. */
+	 no more states can be reached with better costs. */
       {
 	/* XXX - dynamic ringbuffer size */
 	tre_tnfa_approx_reach_t *ringbuffer[512];
@@ -497,7 +507,6 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
 	    tre_tnfa_approx_reach_t *reach_p;
 	    int id;
 	    int depth;
-	    int j;
 	    int cost, cost0;
 	    tre_tnfa_transition_t *trans;
 
@@ -564,9 +573,10 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
 		    && (cost0 > reach_next[dest_id].costs[0][TRE_M_COST]
 			|| (cost0 == reach_next[dest_id].costs[0][TRE_M_COST]
 			    && (!match_tags
-				|| tre_tag_order(num_tags, tnfa->tag_directions,
-						 reach_next[dest_id].tags,
-						 tmp_tags)))))
+				|| !tre_tag_order(num_tags,
+						  tnfa->tag_directions,
+						  tmp_tags,
+						  reach_next[dest_id].tags)))))
 		  {
 		    DPRINT(("lose, cost0 %d, have %d\n",
 			    cost0, reach_next[dest_id].costs[0][TRE_M_COST]));
@@ -586,9 +596,10 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
 				 default_params);
 
 		reach_next[dest_id].depth = reach_p->depth;
-		for (i = 0; i <= depth; i++)
-		  for (j = 0; j < TRE_M_LAST; j++)
-		    reach_next[dest_id].costs[i][j] = reach_p->costs[i][j];
+		memcpy(&reach_next[dest_id].costs,
+		       reach_p->costs,
+		       sizeof(reach_p->costs[0][0])
+		       * TRE_M_LAST * (depth + 1));
 		reach_next[dest_id].costs[depth][TRE_M_COST] = cost;
 		reach_next[dest_id].costs[depth][TRE_M_NUM_DEL]++;
 		reach_next[dest_id].costs[depth][TRE_M_NUM_ERR]++;
@@ -606,11 +617,11 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
 			    && (num_tags > 0
 				&& tmp_tags[0] <= match_tags[0]))))
 		  {
-		    DPRINT(("    setting new match at %d, cost %d\n",
+		    DPRINT(("	 setting new match at %d, cost %d\n",
 			    pos, cost0));
 		    match_eo = pos;
-		    for (i = 0; i < TRE_M_LAST; i++)
-		      match_costs[i] = reach_next[dest_id].costs[0][i];
+		    memcpy(match_costs, reach_next[dest_id].costs[0],
+			   sizeof(match_costs[0]) * TRE_M_LAST);
 		    for (i = 0; i < num_tags; i++)
 		      match_tags[i] = tmp_tags[i];
 		  }
@@ -633,19 +644,20 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
       tre_print_reach(tnfa, reach_next, pos, num_tags);
 #endif /* TRE_DEBUG */
 
-      GET_NEXT_WCHAR();
-
       /* Check for end of string. */
       if (len < 0)
-        {
-          if (prev_c == L'\0')
-            break;
-        }
+	{
+	  if (next_c == L'\0')
+	    break;
+	}
       else
-        {
-          if (pos > len)
-            break;
-        }
+	{
+	  if (pos >= len)
+	    break;
+	}
+
+      prev_pos = pos;
+      GET_NEXT_WCHAR();
 
       /* Swap `reach' and `reach_next'. */
       {
@@ -660,13 +672,12 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
 	{
 	  tre_tnfa_transition_t *trans;
 
-	  /* XXX - `pos - 1' does not work for mbs. */
-	  if (reach[id].pos < pos - 1)
+	  if (reach[id].pos < prev_pos)
 	    continue;  /* Not reached. */
 	  for (trans = reach[id].state; trans->state; trans++)
 	    {
 	      int dest_id;
-	      int j, depth;
+	      int depth;
 	      int cost, cost0, err;
 
 	      if (trans->assertions
@@ -744,16 +755,17 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
 		tmp_tags[i] = reach[id].tags[i];
 	      if (trans->tags)
 		for (i = 0; trans->tags[i] >= 0; i++)
-		  tmp_tags[trans->tags[i]] = pos;
+		  if (trans->tags[i] < num_tags)
+		    tmp_tags[trans->tags[i]] = pos;
 
 	      /* If another path has also reached this state, choose the
 		 one with the smallest cost or best tags if costs are equal. */
 	      if (reach_next[dest_id].pos == pos
 		  && (cost0 > reach_next[dest_id].costs[0][TRE_M_COST]
 		      || (cost0 == reach_next[dest_id].costs[0][TRE_M_COST]
-			  && tre_tag_order(num_tags, tnfa->tag_directions,
-					   reach_next[dest_id].tags,
-					   tmp_tags))))
+			  && !tre_tag_order(num_tags, tnfa->tag_directions,
+					    tmp_tags,
+					    reach_next[dest_id].tags))))
 		{
 		  DPRINT(("lose\n"));
 		  continue;
@@ -776,9 +788,10 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
 			       default_params);
 
 	      /* Set the costs after this transition. */
-	      for (i = 0; i <= depth; i++)
-		for (j = 0; j < TRE_M_LAST; j++)
-		  reach_next[dest_id].costs[i][j] = reach[id].costs[i][j];
+		memcpy(&reach_next[dest_id].costs,
+		       reach[id].costs,
+		       sizeof(reach[id].costs[0][0])
+		       * TRE_M_LAST * (depth + 1));
 	      reach_next[dest_id].costs[depth][TRE_M_COST] = cost;
 	      reach_next[dest_id].costs[depth][TRE_M_NUM_SUBST] += err;
 	      reach_next[dest_id].costs[depth][TRE_M_NUM_ERR] += err;
@@ -809,6 +822,12 @@ tre_tnfa_run_approx(const tre_tnfa_t *tnfa, const void *string, int len,
 
   DPRINT(("match end offset = %d, match cost = %d\n", match_eo,
 	  match_costs[TRE_M_COST]));
+
+#ifndef TRE_USE_ALLOCA
+  if (buf)
+    xfree(buf);
+#endif /* !TRE_USE_ALLOCA */
+
   match->cost = match_costs[TRE_M_COST];
   match->num_ins = match_costs[TRE_M_NUM_INS];
   match->num_del = match_costs[TRE_M_NUM_DEL];
