@@ -1,20 +1,22 @@
 /*
   tre-parse.c - Regexp parser
 
-  Copyright (C) 2001-2004 Ville Laurikari <vl@iki.fi>
+  Copyright (c) 2001-2006 Ville Laurikari <vl@iki.fi>
 
-  This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU General Public License version 2 (June
-  1991) as published by the Free Software Foundation.
+  This library is free software; you can redistribute it and/or
+  modify it under the terms of the GNU Lesser General Public
+  License as published by the Free Software Foundation; either
+  version 2.1 of the License, or (at your option) any later version.
 
-  This program is distributed in the hope that it will be useful,
+  This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-  GNU General Public License for more details.
+  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+  Lesser General Public License for more details.
+
+  You should have received a copy of the GNU Lesser General Public
+  License along with this library; if not, write to the Free Software
+  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-  You should have received a copy of the GNU General Public License
-  along with this program; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
 /*
@@ -632,7 +634,7 @@ tre_parse_bound(tre_parse_ctx_t *ctx, tre_ast_node_t **result)
   int limit_ins, limit_del, limit_subst, limit_err;
   const tre_char_t *r = ctx->re;
   const tre_char_t *start;
-  int minimal = 0;
+  int minimal = (ctx->cflags & REG_UNGREEDY) ? 1 : 0;
   int approx = 0;
   int costs_set = 0;
   int counts_set = 0;
@@ -843,7 +845,7 @@ tre_parse_bound(tre_parse_ctx_t *ctx, tre_ast_node_t **result)
   /* Parse trailing '?' marking minimal repetition. */
   if (r < ctx->re_end && *r == CHAR_QUESTIONMARK)
     {
-      minimal = 1;
+      minimal = !(ctx->cflags & REG_UNGREEDY);
       r++;
     }
 
@@ -1129,21 +1131,28 @@ tre_parse(tre_parse_ctx_t *ctx)
 	    case CHAR_QUESTIONMARK:
 	      {
 		tre_ast_node_t *tmp_node;
-		int minimal = 0;
+		int minimal = (ctx->cflags & REG_UNGREEDY) ? 1 : 0;
 		int rep_min = 0;
 		int rep_max = -1;
+		const tre_char_t *tmp_re;
+
 		if (*ctx->re == CHAR_PLUS)
 		  rep_min = 1;
 		if (*ctx->re == CHAR_QUESTIONMARK)
 		  rep_max = 1;
-
+		tmp_re = ctx->re;
+		
 		if (ctx->re + 1 < ctx->re_end
 		    && *(ctx->re + 1) == CHAR_QUESTIONMARK)
-		  minimal = 1;
+		  { 
+		    minimal = !(ctx->cflags & REG_UNGREEDY);
+		    ctx->re++;
+		  }
+
 		DPRINT(("tre_parse: %s star: '%.*" STRF "'\n",
 			minimal ? "  minimal" : "greedy",
-			ctx->re_end - ctx->re, ctx->re));
-		ctx->re += minimal + 1;
+			ctx->re_end - tmp_re, tmp_re));
+		ctx->re++;
 		tmp_node = tre_ast_new_iter(ctx->mem, result, rep_min, rep_max,
 					    minimal);
 		if (tmp_node == NULL)
@@ -1245,6 +1254,18 @@ tre_parse(tre_parse_ctx_t *ctx)
 			  ctx->re++;
 			}
 #endif /* REG_RIGHT_ASSOC */
+#ifdef REG_UNGREEDY
+		      else if (*ctx->re == L'U')
+			{
+			  DPRINT(("tre_parse:    ungreedy: '%.*" STRF "\n",
+				  ctx->re_end - ctx->re, ctx->re));
+			  if (bit)
+			    new_cflags |= REG_UNGREEDY;
+			  else
+			    new_cflags &= ~REG_UNGREEDY;
+			  ctx->re++;
+			}
+#endif /* REG_UNGREEDY */
 		      else if (*ctx->re == CHAR_MINUS)
 			{
 			  DPRINT(("tre_parse:	 turn off: '%.*" STRF "\n",
@@ -1260,6 +1281,23 @@ tre_parse(tre_parse_ctx_t *ctx)
 			  depth++;
 			  break;
 			}
+		      else if (*ctx->re == CHAR_HASH)
+			{
+			  DPRINT(("tre_parse:    comment: '%.*" STRF "\n",
+				  ctx->re_end - ctx->re, ctx->re));
+			  /* A comment can contain any character except a 
+			     right parenthesis */
+			  while (*ctx->re != CHAR_RPAREN
+				 && ctx->re < ctx->re_end)
+			    ctx->re++;
+			  if (*ctx->re == CHAR_RPAREN && ctx->re < ctx->re_end)
+			    {
+			      ctx->re++;
+			      break;
+			    }
+			  else
+			    return REG_BADPAT;
+			}
 		      else if (*ctx->re == CHAR_RPAREN)
 			{
 			  ctx->re++;
@@ -1590,7 +1628,7 @@ tre_parse(tre_parse_ctx_t *ctx)
 		  ctx->cflags &= ~temporary_cflags;
 		  temporary_cflags = 0;
 		  ctx->re += 2;
-		  STACK_PUSHX(stack, PARSE_ATOM);
+		  STACK_PUSHX(stack, PARSE_PIECE);
 		  break;
 		}
 
