@@ -1,20 +1,21 @@
 /*
   retest.c - TRE regression test program
 
-  Copyright (C) 2001-2004 Ville Laurikari <vl@iki.fi>.
+  Copyright (c) 2001-2006 Ville Laurikari <vl@iki.fi>.
 
-  This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU General Public License version 2 (June
-  1991) as published by the Free Software Foundation.
+  This software is free; you can redistribute it and/or modify it
+  under the terms of the GNU Lesser General Public License as
+  published by the Free Software Foundation; either version 2.1 of the
+  License, or (at your option) any later version.
 
-  This program is distributed in the hope that it will be useful,
+  This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-  GNU General Public License for more details.
+  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+  Lesser General Public License for more details.
 
-  You should have received a copy of the GNU General Public License
-  along with this program; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+  You should have received a copy of the GNU Lesser General Public
+  License along with this software; if not, write to the Free Software
+  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 */
 
@@ -40,15 +41,25 @@
 #endif /* HAVE_MALLOC_H */
 #include <regex.h>
 
+#ifdef TRE_VERSION
+#define HAVE_REGNEXEC 1
+#define HAVE_REGNCOMP 1
 #include "xmalloc.h"
+#else /* !TRE_VERSION */
+#define xmalloc malloc
+#define xfree free
+#endif /* !TRE_VERSION */
 
-#define elementsof(x)   (sizeof(x)/sizeof(x[0]))
+#define elementsof(x)	(sizeof(x)/sizeof(x[0]))
 
 static int valid_reobj = 0;
 static regex_t reobj;
 static regmatch_t pmatch[32];
 static char *regex_pattern;
 static int cflags;
+static int use_regnexec = 0;
+static int use_regncomp = 0;
+static int avoid_eflags = 0;
 
 static int comp_tests = 0;
 static int exec_tests = 0;
@@ -61,42 +72,99 @@ static int exec_errors = 0;
 
 #define END -2
 
+void
+test_status(char c)
+{
+  static int k = 0;
+  printf("%c", c);
+  if (++k % 79 == 0)
+    printf("\n");
+  fflush(stdout);
+}
+
+
+static int
+wrap_regexec(const char *data, size_t len,
+	     size_t pmatch_len, regmatch_t *pmatch, int eflags)
+{
+  char *buf = NULL;
+  int result;
+
+  if (len == 0 && use_regnexec)
+    {
+      /* Zero length string and using regnexec(), the pointer we give
+	 should not be dereferenced at all. */
+      buf = NULL;
+    }
+  else
+    {
+      /* Copy the data to a separate buffer to make a better test for
+	 regexec() and regnexec(). */
+      buf = xmalloc(len + !use_regnexec);
+      if (!buf)
+	return REG_ESPACE;
+      memcpy(buf, data, len);
+      test_status('#');
+    }
+
+#ifdef HAVE_REGNEXEC
+  if (use_regnexec)
+    {
+      if (len == 0)
+	result = regnexec(&reobj, NULL, len, pmatch_len, pmatch, eflags);
+      else
+	result = regnexec(&reobj, buf, len, pmatch_len, pmatch, eflags);
+    }
+  else
+#endif /* HAVE_REGNEXEC */
+    {
+      buf[len] = '\0';
+      result = regexec(&reobj, buf, pmatch_len, pmatch, eflags);
+    }
+
+  xfree(buf);
+  return result;
+}
+
+static int
+wrap_regcomp(regex_t *preg, const char *data, size_t len, int cflags)
+{
+#ifdef HAVE_REGNCOMP
+  if (use_regncomp)
+    return regncomp(preg, data, len, cflags);
+  else
+    return regcomp(preg, data, cflags);
+#else /* !HAVE_REGNCOMP */
+  fprintf(stderr, "%s\n", data);
+  return regcomp(preg, data, cflags);
+#endif /* !HAVE_REGNCOMP */
+}
+
 static int
 execute(char *data, int len, size_t pmatch_len, regmatch_t *pmatch, int eflags)
 {
-#if defined(MALLOC_DEBUGGING) && !defined(TRE_USE_ALLOCA)
-  static int j = 0, k = 0;
+#ifdef MALLOC_DEBUGGING
   int i = 0;
   int ret;
 
   while (1)
     {
-      if (j++ % 20 == 0)
-	{
-	  printf(",");
-	  if (++k % 79 == 0)
-	    printf("\n");
-	  fflush(stdout);
-	}
       xmalloc_configure(i);
       comp_tests++;
-      ret = regnexec(&reobj, data, len, pmatch_len, pmatch, eflags);
+      ret = wrap_regexec(data, len, pmatch_len, pmatch, eflags);
       if (ret != REG_ESPACE)
 	{
-	  printf("#");
-	  if (++k % 79 == 0)
-	    printf("\n");
-	    break;
+	  break;
 	}
 #ifdef REGEX_DEBUG
       xmalloc_dump_leaks();
 #endif /* REGEX_DEBUG */
       i++;
-      return ret;
     }
-#else
-  return regnexec(&reobj, data, len, pmatch_len, pmatch, eflags);
-#endif
+  return ret;
+#else /* !MALLOC_DEBUGGING */
+  return wrap_regexec(data, len, pmatch_len, pmatch, eflags);
+#endif /* !MALLOC_DEBUGGING */
 }
 
 static int
@@ -110,13 +178,13 @@ check(va_list ap, int ret, char *str, size_t pmatch_len, regmatch_t *pmatch,
       printf("Exec error, regex: \"%s\", cflags %d, "
 	     "string: \"%s\", eflags %d\n", regex_pattern, cflags,
 	     str, eflags);
-      printf("  got %smatch (regnexec returned %d)\n", ret ? "no " : "", ret);
-      fail = 1;
+      printf("	got %smatch (regexec returned %d)\n", ret ? "no " : "", ret);
+      return 1;
     }
 
-  if (!fail && ret == 0)
+  if (ret == 0)
     {
-      int i;
+      unsigned int i;
 
       for (i = 0; i < pmatch_len; i++)
 	{
@@ -130,7 +198,7 @@ check(va_list ap, int ret, char *str, size_t pmatch_len, regmatch_t *pmatch,
 	    {
 	      printf("Exec error, regex: \"%s\", string: \"%s\"\n",
 		     regex_pattern, str);
-	      printf("  group %d: expected (%d, %d) \"%.*s\", "
+	      printf("	group %d: expected (%d, %d) \"%.*s\", "
 		     "got (%d, %d) \"%.*s\"\n",
 		     i, rm_so, rm_eo, rm_eo - rm_so, str + rm_so,
 		     pmatch[i].rm_so, pmatch[i].rm_eo,
@@ -166,13 +234,13 @@ check(va_list ap, int ret, char *str, size_t pmatch_len, regmatch_t *pmatch,
 
 
 static void
-test_nexec_do(char *str, size_t len, int eflags, va_list ap)
+test_nexec(char *data, size_t len, int eflags, ...)
 {
   int m;
   int fail = 0;
-  char *data;
-
-  exec_tests++;
+  int extra_flags[] = {0, REG_BACKTRACKING_MATCHER, REG_APPROX_MATCHER};
+  int i;
+  va_list ap;
 
   if (!valid_reobj)
     {
@@ -180,73 +248,98 @@ test_nexec_do(char *str, size_t len, int eflags, va_list ap)
       return;
     }
 
-#ifdef MALLOC_DEBUGGING
-  xmalloc_configure(1);
-#endif /* !MALLOC_DEBUGGING */
+  use_regnexec = 1;
 
-  if (len == 0)
+  for (i = 0; i < elementsof(extra_flags); i++)
     {
-      data = NULL;
+      int final_flags = eflags | extra_flags[i];
+
+      if ((final_flags & REG_BACKTRACKING_MATCHER
+	   && tre_have_approx(&reobj))
+	  || (final_flags & REG_APPROX_MATCHER
+	      && tre_have_backrefs(&reobj))
+	  || (final_flags & avoid_eflags))
+	continue;
+
+      /* Test with a pmatch array. */
+      exec_tests++;
+      m = execute(data, len, elementsof(pmatch), pmatch, final_flags);
+      va_start(ap, eflags);
+      fail |= check(ap, m, data, elementsof(pmatch), pmatch, final_flags);
+      va_end(ap);
+
+      /* Same test with a NULL pmatch. */
+      exec_tests++;
+      m = execute(data, len, 0, NULL, final_flags);
+      va_start(ap, eflags);
+      fail |= check(ap, m, data, 0, NULL, final_flags);
+      va_end(ap);
     }
-  else
-    {
-      data = xmalloc(len);
-      if (data == NULL)
-	{
-	  fprintf(stderr, "Out of memory.\n");
-	  exit(1);
-	}
-      memcpy(data, str, len);
-    }
-#ifdef MALLOC_DEBUGGING
-  xmalloc_configure(0);
-#endif /* !MALLOC_DEBUGGING */
-
-  /* XXX - test the approximate matcher (with cost 0) and the
-     backtracking matcher as well using special eflags. */
-
-  /* Test with a pmatch array. */
-  m = execute(data, len, elementsof(pmatch), pmatch, eflags);
-  fail = check(ap, m, str, elementsof(pmatch), pmatch, eflags);
-  va_end(ap);
-
-  /* Same test with a NULL pmatch. */
-  m = execute(data, len, 0, NULL, eflags);
-  fail |= check(ap, m, str, 0, NULL, eflags);
-  va_end(ap);
-
-  xfree(data);
 
   if (fail)
     exec_errors++;
 }
 
+
+
 static void
 test_exec(char *str, int eflags, ...)
 {
+  int m;
+  int fail = 0;
+  size_t len = strlen(str);
+  int extra_flags[] = {0,
+		       REG_BACKTRACKING_MATCHER,
+		       REG_APPROX_MATCHER,
+		       REG_BACKTRACKING_MATCHER | REG_APPROX_MATCHER};
+  int i;
   va_list ap;
 
-  va_start(ap, eflags);
-  test_nexec_do(str, strlen(str), eflags, ap);
-  va_end(ap);
+  if (!valid_reobj)
+    {
+      exec_errors++;
+      return;
+    }
+
+  for (use_regnexec = 0; use_regnexec < 2; use_regnexec++)
+    {
+      for (i = 0; i < elementsof(extra_flags); i++)
+	{
+	  int final_flags = eflags | extra_flags[i];
+
+	  if ((final_flags & REG_BACKTRACKING_MATCHER
+	       && tre_have_approx(&reobj))
+	      || (final_flags & REG_APPROX_MATCHER
+		  && tre_have_backrefs(&reobj))
+	      || (final_flags & avoid_eflags))
+	    continue;
+
+	  /* Test with a pmatch array. */
+	  exec_tests++;
+	  m = execute(str, len, elementsof(pmatch), pmatch, final_flags);
+	  va_start(ap, eflags);
+	  fail |= check(ap, m, str, elementsof(pmatch), pmatch, final_flags);
+	  va_end(ap);
+
+	  /* Same test with a NULL pmatch. */
+	  exec_tests++;
+	  m = execute(str, len, 0, NULL, final_flags);
+	  va_start(ap, eflags);
+	  fail |= check(ap, m, str, 0, NULL, final_flags);
+	  va_end(ap);
+	}
+    }
+
+  if (fail)
+    exec_errors++;
 }
 
-static void
-test_nexec(char *str, size_t len, int eflags, ...)
-{
-  va_list ap;
-
-  va_start(ap, eflags);
-  test_nexec_do(str, len, eflags, ap);
-  va_end(ap);
-}
 
 void
 test_comp(char *re, int flags, int ret)
 {
   int errcode = 0;
-  char *data;
-  int len;
+  int len = strlen(re);
   regex_pattern = re;
   cflags = flags;
 
@@ -258,47 +351,20 @@ test_comp(char *re, int flags, int ret)
       valid_reobj = 0;
     }
 
-
-#ifdef MALLOC_DEBUGGING
-  xmalloc_configure(1);
-#endif /* MALLOC_DEBUGGING */
-  len = strlen(re);
-  if (len == 0)
-    {
-      data = NULL;
-    }
-  else
-    {
-      data = xmalloc(len);
-      if (data == NULL)
-	{
-	  fprintf(stderr, "Out of memory.\n");
-	  exit(1);
-	}
-      memcpy(data, re, len);
-    }
-
 #ifdef MALLOC_DEBUGGING
   {
-    static int j = 0, k = 0;
+    static int j = 0;
     int i = 0;
     while (1)
       {
-	if (j++ % 20 == 0)
-	  {
-	    printf(".");
-	    if (++k % 79 == 0)
-	      printf("\n");
-	    fflush(stdout);
-	  }
 	xmalloc_configure(i);
 	comp_tests++;
-	errcode = regncomp(&reobj, data, len, flags);
+	if (j++ % 20 == 0)
+	  test_status('.');
+	errcode = wrap_regcomp(&reobj, re, len, flags);
 	if (errcode != REG_ESPACE)
 	  {
-	    printf("*");
-	    if (++k % 79 == 0)
-	      printf("\n");
+	    test_status('*');
 	    break;
 	  }
 #ifdef REGEX_DEBUG
@@ -308,19 +374,17 @@ test_comp(char *re, int flags, int ret)
       }
   }
 #else /* !MALLOC_DEBUGGING */
-  errcode = regncomp(&reobj, data, len, flags);
+  errcode = wrap_regcomp(&reobj, re, len, flags);
 #endif /* !MALLOC_DEBUGGING */
 
   if (errcode != ret)
     {
       printf("Comp error, regex: \"%s\"\n", regex_pattern);
-      printf("  expected return code %d, got %d.\n",
+      printf("	expected return code %d, got %d.\n",
 	     ret, errcode);
       comp_errors++;
     }
 
-  xfree(data);
-
   if (errcode == 0)
     valid_reobj = 1;
 }
@@ -329,20 +393,19 @@ test_comp(char *re, int flags, int ret)
 int
 main(int argc, char **argv)
 {
-  test_comp("c$", REG_EXTENDED, 0);
-  test_exec("abc", 0, REG_OK, 2, 3, END);
-
-  test_comp("[A-Z]\\d\\s?\\d[A-Z]{2}", REG_EXTENDED, 0);
-
-  /* Huge memory consumption. */
-  test_comp("[A-Z]\\d\\s?\\d[A-Z]{2}|[A-Z]\\d{2}\\s?\\d[A-Z]{2}|[A-Z]{2}\\d\\s?\\d[A-Z]{2}|[A-Z]{2}\\d{2}\\s?\\d[A-Z]{2}|[A-Z]\\d[A-Z]\\s?\\d[A-Z]{2}|[A-Z]{2}\\d[A-Z]\\s?\\d[A-Z]{2}|[A-Z]{3}\\s?\\d[A-Z]{2}", REG_EXTENDED, 0);
 
+  /* Large number of macros in one regexp. */
+  test_comp("[A-Z]\\d\\s?\\d[A-Z]{2}|[A-Z]\\d{2}\\s?\\d[A-Z]{2}|[A-Z]{2}\\d"
+	    "\\s?\\d[A-Z]{2}|[A-Z]{2}\\d{2}\\s?\\d[A-Z]{2}|[A-Z]\\d[A-Z]\\s?"
+	    "\\d[A-Z]{2}|[A-Z]{2}\\d[A-Z]\\s?\\d[A-Z]{2}|[A-Z]{3}\\s?\\d[A-Z]"
+	    "{2}", REG_EXTENDED, 0);
 
   test_comp("a{11}(b{2}c){2}", REG_EXTENDED, 0);
   test_comp("a{2}{2}xb+xc*xd?x", REG_EXTENDED, 0);
   test_comp("^!packet [0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3} [0-9]+",
 	    REG_EXTENDED, 0);
-  test_comp("^!pfast [0-9]{1,15} ([0-9]{1,3}\\.){3}[0-9]{1,3}[0-9]{1,5}$", REG_EXTENDED, 0);
+  test_comp("^!pfast [0-9]{1,15} ([0-9]{1,3}\\.){3}[0-9]{1,3}[0-9]{1,5}$",
+	    REG_EXTENDED, 0);
 
 #if 0
   /* Should these match or not? */
@@ -483,9 +546,11 @@ main(int argc, char **argv)
   test_comp("(a*)+", REG_EXTENDED, 0);
   test_exec("-", 0, REG_OK, 0, 0, 0, 0, END);
 
+#if 0
   test_comp("((a*)*b)*b", REG_EXTENDED, 0);
   test_exec("aaaaaaaaaaaaaaaaaaaaaaaaab", 0, REG_OK,
 	    25, 26, -1, -1, -1, -1, END);
+#endif
 
   test_comp("", 0, 0);
   test_exec("", 0, REG_OK, 0, 0, END);
@@ -553,6 +618,7 @@ main(int argc, char **argv)
   test_exec("foozot", 0, REG_OK, 0, 6, 3, 3, -1, -1, END);
   test_exec("foobarzot", 0, REG_OK, 0, 9, 3, 6, 3, 6, END);
   test_exec("foobarbarzot", 0, REG_OK, 0, 12, 3, 9, 6, 9, END);
+#if 0
   test_comp("foo((zup)*|(bar)*|(zap)*)*zot", REG_EXTENDED, 0);
   test_exec("foobarzapzot", 0, REG_OK,
 	    0, 12, 6, 9, -1, -1, -1, -1, 6, 9, END);
@@ -566,7 +632,7 @@ main(int argc, char **argv)
 	    0, 9, 3, 6, -1, -1, -1, -1, 3, 6, END);
   test_exec("foozot", 0, REG_OK,
 	    0, 6, 3, 3, -1, -1, -1, -1, -1, -1, END);
-
+#endif
 
 
   /* Test case where, e.g., Perl and Python regexp functions, and many
@@ -637,9 +703,9 @@ main(int argc, char **argv)
 	    "konetaisknoteai", 0, 0);
 
   test_comp("((aab)|(aac)|(aa*))c", REG_EXTENDED, 0);
-  test_exec("aabc", 0, REG_OK, 0, 4, 0, 3,  0,  3, -1, -1, -1, -1, END);
-  test_exec("aacc", 0, REG_OK, 0, 4, 0, 3, -1, -1,  0,  3, -1, -1, END);
-  test_exec("aaac", 0, REG_OK, 0, 4, 0, 3, -1, -1, -1, -1,  0,  3, END);
+  test_exec("aabc", 0, REG_OK, 0, 4, 0, 3,  0,	3, -1, -1, -1, -1, END);
+  test_exec("aacc", 0, REG_OK, 0, 4, 0, 3, -1, -1,  0,	3, -1, -1, END);
+  test_exec("aaac", 0, REG_OK, 0, 4, 0, 3, -1, -1, -1, -1,  0,	3, END);
 
   test_comp("^(([^!]+!)?([^!]+)|.+!([^!]+!)([^!]+))$",
 	    REG_EXTENDED, 0);
@@ -1007,7 +1073,7 @@ main(int argc, char **argv)
   test_comp("^$", 0, 0);
   test_exec("x", 0, REG_NOMATCH);
   test_exec("", 0, REG_OK, 0, 0, END);
-  test_exec("\n", 0, REG_NOMATCH, 0, 0, END);
+  test_exec("\n", 0, REG_NOMATCH);
   test_comp("^$", REG_NEWLINE, 0);
   test_exec("x", 0, REG_NOMATCH);
   test_exec("", 0, REG_OK, 0, 0, END);
@@ -1052,6 +1118,61 @@ main(int argc, char **argv)
   test_comp("\\t", REG_EXTENDED, 0);
   test_comp("\\e", REG_EXTENDED, 0);
 
+  /* Test the \x1B and \x{263a} extensions for specifying 8 bit and wide
+     characters in hexadecimal. */
+  test_comp("\\x41", REG_EXTENDED, 0);
+  test_exec("ABC", 0, REG_OK, 0, 1, END);
+  test_comp("\\x5", REG_EXTENDED, 0);
+  test_exec("\005", 0, REG_OK, 0, 1, END);
+  test_comp("\\x5r", REG_EXTENDED, 0);
+  test_exec("\005r", 0, REG_OK, 0, 2, END);
+  test_comp("\\x", REG_EXTENDED, 0);
+  test_nexec("\000", 1, 0, REG_OK, 0, 1, END);
+  test_comp("\\xr", REG_EXTENDED, 0);
+  test_nexec("\000r", 2, 0, REG_OK, 0, 2, END);
+  test_comp("\\x{41}", REG_EXTENDED, 0);
+  test_exec("ABC", 0, REG_OK, 0, 1, END);
+  test_comp("\\x{5}", REG_EXTENDED, 0);
+  test_exec("\005", 0, REG_OK, 0, 1, END);
+  test_comp("\\x{5}r", REG_EXTENDED, 0);
+  test_exec("\005r", 0, REG_OK, 0, 2, END);
+  test_comp("\\x{}", REG_EXTENDED, 0);
+  test_nexec("\000", 1, 0, REG_OK, 0, 1, END);
+  test_comp("\\x{}r", REG_EXTENDED, 0);
+  test_nexec("\000r", 2, 0, REG_OK, 0, 2, END);
+
+  /* Tests for (?inrU-inrU) and (?inrU-inrU:) */
+  test_comp("foo(?i)bar", REG_EXTENDED, 0);
+  test_exec("fooBaR", 0, REG_OK, 0, 6, END);
+  test_comp("foo(?i)bar|zap", REG_EXTENDED, 0);
+  test_exec("fooBaR", 0, REG_OK, 0, 6, END);
+  test_exec("foozap", 0, REG_OK, 0, 6, END);
+  test_exec("foozAp", 0, REG_OK, 0, 6, END);
+  test_exec("zap", 0, REG_NOMATCH);
+  test_comp("foo(?-i:zap)zot", REG_EXTENDED | REG_ICASE, 0);
+  test_exec("FoOzapZOt", 0, REG_OK, 0, 9, END);
+  test_exec("FoOzApZOt", 0, REG_NOMATCH);
+  test_comp("foo(?i:bar|zap)", REG_EXTENDED, 0);
+  test_exec("foozap", 0, REG_OK, 0, 6, END);
+  test_exec("foobar", 0, REG_OK, 0, 6, END);
+  test_exec("foobAr", 0, REG_OK, 0, 6, END);
+  test_exec("fooZaP", 0, REG_OK, 0, 6, END);
+  test_comp("foo(?U:o*)(o*)", REG_EXTENDED, 0);
+  test_exec("foooo", 0, REG_OK, 0, 5, 3, 5, END);
+
+  /* Test comment syntax. */
+  test_comp("foo(?# This here is a comment. )bar", REG_EXTENDED, 0);
+  test_exec("foobar", 0, REG_OK, 0, 6, END);
+
+  /* Tests for \Q and \E. */
+  test_comp("\\((\\Q)?:\\<[^$\\E)", REG_EXTENDED, 0);
+  test_exec("()?:\\<[^$", 0, REG_OK, 0, 9, 1, 9, END);
+  test_comp("\\Qabc\\E.*", REG_EXTENDED, 0);
+  test_exec("abcdef", 0, REG_OK, 0, 6, END);
+  test_comp("\\Qabc\\E.*|foo", REG_EXTENDED, 0);
+  test_exec("parabc123wxyz", 0, REG_OK, 3, 13, END);
+  test_exec("fooabc123wxyz", 0, REG_OK, 0, 3, END);
+
   /*
    * Test bounded repetitions.
    */
@@ -1078,7 +1199,7 @@ main(int argc, char **argv)
   test_exec("aaa", 0, REG_OK, 0, 3, END);
   test_comp("a{2,}", REG_EXTENDED, REG_OK);
   test_exec("", 0, REG_NOMATCH);
-  test_exec("a", 0, REG_NOMATCH, 0, 1, END);
+  test_exec("a", 0, REG_NOMATCH);
   test_exec("aa", 0, REG_OK, 0, 2, END);
   test_exec("aaa", 0, REG_OK, 0, 3, END);
   test_comp("a{3,}", REG_EXTENDED, REG_OK);
@@ -1163,11 +1284,11 @@ main(int argc, char **argv)
   test_exec("xxxxxxxxxx", 0, REG_OK, 0, 9, 6, 9, 6, 8, END);
   test_exec("xxxxxxxxxxx", 0, REG_OK, 0, 11, 8, 11, 8, 10, END);
   test_comp("a{2}{2}x", REG_EXTENDED, 0);
-  test_exec("", 0, REG_NOMATCH, END);
-  test_exec("x", 0, REG_NOMATCH, END);
-  test_exec("ax", 0, REG_NOMATCH, END);
-  test_exec("aax", 0, REG_NOMATCH, END);
-  test_exec("aaax", 0, REG_NOMATCH, END);
+  test_exec("", 0, REG_NOMATCH);
+  test_exec("x", 0, REG_NOMATCH);
+  test_exec("ax", 0, REG_NOMATCH);
+  test_exec("aax", 0, REG_NOMATCH);
+  test_exec("aaax", 0, REG_NOMATCH);
   test_exec("aaaax", 0, REG_OK, 0, 5, END);
   test_exec("aaaaax", 0, REG_OK, 1, 6, END);
   test_exec("aaaaaax", 0, REG_OK, 2, 7, END);
@@ -1231,8 +1352,10 @@ main(int argc, char **argv)
   test_comp("((.*)\\1)+", REG_EXTENDED, 0);
   test_exec("aa", 0, REG_OK, 0, 2, 0, 2, 0, 1, END);
 
+#if 0
   test_comp("()(\\1\\1)*", REG_EXTENDED, 0);
   test_exec("", 0, REG_OK, 0, 0, 0, 0, -1, -1, END);
+#endif
 
   /* Check that back references work with REG_NOSUB. */
   test_comp("(o)\\1", REG_EXTENDED | REG_NOSUB, 0);
@@ -1242,10 +1365,26 @@ main(int argc, char **argv)
   test_comp("(o)\\1", REG_EXTENDED, 0);
   test_exec("fobar", 0, REG_NOMATCH);
 
+  test_comp("\\1foo", REG_EXTENDED, REG_ESUBREG);
+  test_comp("\\1foo(bar)", REG_EXTENDED, 0);
+
+  /* Back reference with zero-width assertion. */
+  test_comp("(.)\\1$", REG_EXTENDED, 0);
+  test_exec("foox", 0, REG_NOMATCH);
+  test_exec("foo", 0, REG_OK, 1, 3, 1, 2, END);
+
+  /* Back references together with {}. */
+  test_comp("([0-9]{5})\\1", REG_EXTENDED, 0);
+  test_exec("12345", 0, REG_NOMATCH);
+  test_exec("1234512345", 0, REG_OK, 0, 10, 0, 5, END);
+  test_comp("([0-9]{4})\\1", REG_EXTENDED, 0);
+  test_exec("1234", 0, REG_NOMATCH);
+  test_exec("12341234", 0, REG_OK, 0, 8, 0, 4, END);
 
   /*
    * Test minimal repetitions (non-greedy repetitions)
    */
+  avoid_eflags = REG_BACKTRACKING_MATCHER | REG_APPROX_MATCHER;
 
   /* Basic .*/
   test_comp(".*?", REG_EXTENDED, 0);
@@ -1264,6 +1403,14 @@ main(int argc, char **argv)
   test_exec("hubba wooga-booga zabar gafoo wazap", 0, REG_OK,
 	    4, 23, 5, 20, 20, 23, END);
 
+  /* Test REG_UNGREEDY. */
+  test_comp(".*", REG_EXTENDED | REG_UNGREEDY, 0);
+  test_exec("abcd", 0, REG_OK, 0, 0, END);
+  test_comp(".*?", REG_EXTENDED | REG_UNGREEDY, 0);
+  test_exec("abcd", 0, REG_OK, 0, 4, END);
+
+  avoid_eflags = 0;
+  
 
   /*
    * Error reporting tests.
@@ -1318,8 +1465,8 @@ main(int argc, char **argv)
     {
       printf("\nTesting LC_CTYPE ja_JP.eucjp\n");
       /* I tried to make a test where implementations not aware of multibyte
-         character sets will fail.  I have no idea what the japanese text here
-         means, I took it from http://www.ipsec.co.jp/. */
+	 character sets will fail.  I have no idea what the japanese text here
+	 means, I took it from http://www.ipsec.co.jp/. */
       test_comp("+", REG_EXTENDED, 0);
       test_exec("ξޤϡ", 0, REG_OK, 10, 12, END);
 
@@ -1327,10 +1474,9 @@ main(int argc, char **argv)
       test_nexec("foo\000bar", 7, 0, REG_OK, 5, 6, END);
       test_comp("c$", REG_EXTENDED, 0);
       test_exec("abc", 0, REG_OK, 2, 3, END);
-
     }
 #endif /* TRE_MULTIBYTE */
-#endif /* WIN32 */
+#endif /* !WIN32 */
 
   regfree(&reobj);
 
